# -*- coding:utf-8 -*-
"""
Author: fandi.di@bytedance.com
Reference:
    [1]
"""

from ..feature_column import build_input_features, input_from_feature_columns
from ..layers import combined_dnn_input, DNN, PredictionLayer
from ..layers.multi_task import MMoELayer
import tensorflow as tf


def MMOE(dnn_feature_columns, num_tasks, tasks, tasks_name, num_experts=4, units_experts=128, task_dnn_units=(32, 32),
         seed=1024, dnn_activation='relu'):
    """Instantiates the Multi-gate Mixture-of-Experts architecture.

    Args:
        dnn_feature_columns: An iterable containing all the features used by deep part of the model.
        num_tasks: integer, number of tasks, equal to number of outputs, must be greater than 1.
        tasks: list of str, indicating the loss of each tasks, ``"binary"`` for  binary logloss, ``"regression"`` for regression loss. e.g. ['binary', 'regression']
        num_experts: integer, number of experts.
        units_experts: integer, the hidden units of each expert.
        tasks_name: list of str, the name of each tasks,
        task_dnn_units: list,list of positive integer or empty list, the layer number and units in each layer of task-specific DNN
        dnn_activation: Activation function to use in DNN
    return: return a Keras model instance.
    """
    if num_tasks <= 1:
        raise ValueError("num_tasks must be greater than 1")
    if len(tasks) != num_tasks:
        raise ValueError("num_tasks must be equal to the length of tasks")
    for task in tasks:
        if task not in ['binary', 'regression']:
            raise ValueError("task must be binary or regression, {} is illegal".format(task))
    # 特征输入

    features = build_input_features(dnn_feature_columns)
    inputs_list = list(features.values())

    dnn_sparse_embedding_list, dnn_dense_value_list = input_from_feature_columns(features, dnn_feature_columns,
                                                                         l2_reg = 1e-5, seed=seed)

    dnn_input = combined_dnn_input(dnn_sparse_embedding_list, dnn_dense_value_list)

    # MMOELayer
    mmoe_layers = MMoELayer(units_experts=units_experts, num_tasks=num_tasks, num_experts=num_experts,
                            name='mmoe_layer')(dnn_input)

    # 分别处理不同 Task Tower
    task_outputs = []
    for task_layer, task, task_name in zip(mmoe_layers, tasks, tasks_name):
        tower_layer = DNN(hidden_units=task_dnn_units, activation='relu', name='tower_{}'.format(task_name))(task_layer)

        # batch_size * 1
        output_layer = tf.keras.layers.Dense(units=1, activation=None, use_bias=False,
                                             kernel_initializer=tf.keras.initializers.VarianceScaling(),
                                             name='logit_{}'.format(task_name))(tower_layer)
        output = PredictionLayer(task, name=task_name)(output_layer)

        task_outputs.append(output)

    model = tf.keras.models.Model(inputs=inputs_list, outputs=task_outputs)

    return model